@@ -1,23 +1,47 @@
-
 import os
-import z3
 import ctypes
 import logging
 import numbers
 import operator
 import threading
 import weakref
+import signal
 from functools import reduce
 from decimal import Decimal
+import z3
 
 from cachetools import LRUCache
 
-from ..errors import ClaripyZ3Error
+from ..errors import ClaripyZ3Error, ClaripySolverInterruptError
 
 l = logging.getLogger("claripy.backends.backend_z3")
 
 #pylint:disable=unidiomatic-typecheck
 
+ALL_Z3_CONTEXTS = weakref.WeakSet()
+def handle_sigint(signals, frametype):
+    if old_handler == signal.SIG_IGN:
+        return
+
+    contexts = list(ALL_Z3_CONTEXTS)
+    for context in contexts:
+        context.interrupt()
+
+    if old_handler is signal.default_int_handler:
+        raise KeyboardInterrupt()
+    if callable(old_handler):
+        old_handler(signals, frametype)
+    else:
+        print("*** CRITICAL ERROR - THIS SHOULD NEVER HAPPEN")
+        raise KeyboardInterrupt()
+
+old_handler = signal.getsignal(signal.SIGINT)
+if old_handler is None or old_handler == signal.SIG_DFL:
+    # there is a signal handler installed by someone other than python. we cannot handle this.
+    old_handler = True
+else:
+    signal.signal(signal.SIGINT, handle_sigint)
+
 try:
     import __pypy__
     _is_pypy = True
@@ -25,9 +49,9 @@ except ImportError:
     _is_pypy = False
 
 def z3_expr_to_smt2(f, status="unknown", name="benchmark", logic=""):
-      # from https://stackoverflow.com/a/14629021/9719920
-      v = (z3.Ast * 0)()
-      return z3.Z3_benchmark_to_smtlib_string(f.ctx_ref(), name, logic, status, "", 0, v, f.as_ast())
+    # from https://stackoverflow.com/a/14629021/9719920
+    v = (z3.Ast * 0)()
+    return z3.Z3_benchmark_to_smtlib_string(f.ctx_ref(), name, logic, status, "", 0, v, f.as_ast())
 
 def claripy_solver_to_smt2(s):
     return s._get_solver().to_smt2()
@@ -75,6 +99,20 @@ def _z3_decl_name_str(ctx, decl):
     symbol_name = z3.Z3_get_symbol_string_bytes(ctx, decl_name)
     return symbol_name
 
+def z3_solver_sat(solver, extra_constraints, occasion):
+    l.debug("Doing a check! (%s)", occasion)
+
+    result = solver.check(extra_constraints)
+
+    if result == z3.unknown:
+        reason = solver.reason_unknown()
+        if reason in ('interrupted from keyboard', 'interrupted'):
+            raise KeyboardInterrupt()
+        elif reason in ('timeout', 'max. resource limit exceeded', 'max. memory exceeded'):
+            raise ClaripySolverInterruptError(reason)
+        else:
+            raise ClaripyZ3Error('solver unknown: ' + reason)
+    return result == z3.sat
 
 class SmartLRUCache(LRUCache):
     def __init__(self, maxsize, getsizeof=None, evict=None):
@@ -174,7 +212,9 @@ class BackendZ3(Backend):
         try:
             return self._tls.context
         except AttributeError:
-            self._tls.context = z3.Context() if threading.current_thread().name != 'MainThread' else z3.main_ctx()
+            main_thread = threading.current_thread() == threading.main_thread()
+            self._tls.context = z3.Context() if not main_thread else z3.main_ctx()
+            ALL_Z3_CONTEXTS.add(self._tls.context)
             return self._tls.context
 
     @property
@@ -674,16 +714,18 @@ class BackendZ3(Backend):
         # see https://stackoverflow.com/a/58829514/3154996
         if z3.Z3_get_decl_num_parameters(ctx, decl) != 1:
             raise BackendError("Weird Z3 model")
-        if z3.Z3_get_decl_parameter_kind(ctx, decl, 0) != z3.Z3_PK_SYMBOL:
+        if z3.Z3_get_decl_parameter_kind(ctx, decl, 0) != z3.Z3_PARAMETER_SYMBOL:
             raise BackendError("Weird Z3 model")
         symb = z3.Z3_get_decl_symbol_parameter(ctx, decl, 0)
         if z3.Z3_get_symbol_kind(ctx, symb) != z3.Z3_STRING_SYMBOL:
             raise BackendError("Weird Z3 model")
         return z3.Z3_get_symbol_string(ctx, symb).encode().decode('unicode_escape')
 
-    def solver(self, timeout=None):
+    def solver(self, timeout=None, max_memory=None):
         if not self.reuse_z3_solver or getattr(self._tls, 'solver', None) is None:
             s = z3.Solver(ctx=self._context)
+            if threading.current_thread() != threading.main_thread():
+                s.set(ctrl_c=False)
             _add_memory_pressure(1024 * 1024 * 10)
             if self.reuse_z3_solver:
                 # Store the Z3 solver to a thread-local storage if the reuse-solver option is enabled
@@ -700,8 +742,16 @@ class BackendZ3(Backend):
                 s.set('solver2_timeout', timeout)
             else:
                 s.set('timeout', timeout)
+        if max_memory is not None:
+            s.set('max_memory', max_memory)
         return s
 
+    def clone_solver(self, s):
+        # This clones the solver.
+        # See https://github.com/Z3Prover/z3/issues/556
+        clone = s.translate(self._context)
+        return clone
+
     def _add(self, s, c, track=False):
         if track:
             already_tracked = set(str(impl.children()[0]) for impl in s.assertions())
@@ -709,6 +759,7 @@ class BackendZ3(Backend):
                 name = str(hash(constraint))
                 if name not in already_tracked:
                     s.assert_and_track(constraint, name)
+                    already_tracked.add(name)
         else:
             s.add(*c)
 
@@ -751,21 +802,13 @@ class BackendZ3(Backend):
         global solve_count
 
         solve_count += 1
-        if len(extra_constraints) > 0:
-            solver.push()
-            solver.add(*extra_constraints)
 
-        try:
+        l.debug("Doing a check! (satisfiable)")
+        if not z3_solver_sat(solver, extra_constraints, "satisfiable"):
+            return False
 
-            l.debug("Doing a check! (satisfiable)")
-            if solver.check() != z3.sat:
-                return False
-
-            if model_callback is not None:
-                model_callback(self._generic_model(solver.model()))
-        finally:
-            if len(extra_constraints) > 0:
-                solver.pop()
+        if model_callback is not None:
+            model_callback(self._generic_model(solver.model()))
         return True
 
     def _eval(self, expr, n, extra_constraints=(), solver=None, model_callback=None):
@@ -783,15 +826,12 @@ class BackendZ3(Backend):
 
         result_values = [ ]
 
-        if len(extra_constraints) > 0 or n != 1:
+        if n > 1:
             solver.push()
-        if len(extra_constraints) > 0:
-            solver.add(*extra_constraints)
 
         for i in range(n):
             solve_count += 1
-            l.debug("Doing a check! (batch_eval)")
-            if solver.check() != z3.sat:
+            if not z3_solver_sat(solver, extra_constraints, "batch_eval"):
                 break
             model = solver.model()
 
@@ -817,7 +857,7 @@ class BackendZ3(Backend):
                     solver.add(self._op_raw_Not(self._op_raw_And(*[(ex == ex_v) for ex, ex_v in zip(exprs, r)])))
                 model = None
 
-        if len(extra_constraints) > 0 or n != 1:
+        if n > 1:
             solver.pop()
 
         return result_values
@@ -833,13 +873,9 @@ class BackendZ3(Backend):
             lo = 0
             hi = 2**expr.size()-1
 
-        vals = set()
+        extra_constraints_converted = [self.convert(e) for e in extra_constraints]
+        new_constraints = []
 
-        if len(extra_constraints) > 0:
-            solver.push()
-            solver.add(*[self.convert(e) for e in extra_constraints])
-
-        numpop = 0
         GE = operator.ge if signed else z3.UGE
         LE = operator.le if signed else z3.ULE
 
@@ -848,51 +884,36 @@ class BackendZ3(Backend):
             middle = (lo + hi)//2
             #l.debug("h/m/l/d: %d %d %d %d", hi, middle, lo, hi-lo)
 
-            solver.push()
             # TODO: is this assumption correct?
             # here it's not safe to call directly the z3 low level API since it might happen that the argument is an
             # integer and not a BV
-            solver.add(GE(expr, lo), LE(expr, middle))
-            numpop += 1
+            new_constraints.append(z3.And(GE(expr, lo), LE(expr, middle)))
 
             solve_count += 1
-            l.debug("Doing a check! (min)")
-            if solver.check() == z3.sat:
+            if z3_solver_sat(solver, extra_constraints_converted + new_constraints, "min"):
                 l.debug("... still sat")
                 if model_callback is not None:
                     model_callback(self._generic_model(solver.model()))
-                vals.add(self._primitive_from_model(solver.model(), expr))
                 hi = middle
             else:
                 l.debug("... now unsat")
                 lo = middle
-                solver.pop()
-                numpop -= 1
-
-        for _ in range(numpop):
-            solver.pop()
+                new_constraints.pop()
 
         #l.debug("final hi/lo: %d, %d", hi, lo)
 
+        ret = None
         if hi == lo:
-            vals.add(lo)
+            ret = lo
         else:
-            solver.push()
-            solver.add(expr == lo)
-            l.debug("Doing a check! (min)")
-            if solver.check() == z3.sat:
+            if z3_solver_sat(solver, extra_constraints_converted + [expr == lo], "min"):
                 if model_callback is not None:
                     model_callback(self._generic_model(solver.model()))
-                vals.add(lo)
-                solver.pop()
+                ret = lo
             else:
-                vals.add(hi)
-                solver.pop()
+                ret = hi
 
-        if len(extra_constraints) > 0:
-            solver.pop()
-
-        return min(vals)
+        return ret
 
     @condom
     def _max(self, expr, extra_constraints=(), signed=False, solver=None, model_callback=None):
@@ -904,13 +925,10 @@ class BackendZ3(Backend):
         else:
             lo = 0
             hi = 2**expr.size()-1
-        vals = set()
 
-        if len(extra_constraints) > 0:
-            solver.push()
-            solver.add(*[self.convert(e) for e in extra_constraints])
+        extra_constraints_converted = [self.convert(e) for e in extra_constraints]
+        new_constraints = []
 
-        numpop = 0
         GT = operator.gt if signed else z3.UGT
         LE = operator.le if signed else z3.ULE
 
@@ -919,50 +937,35 @@ class BackendZ3(Backend):
             middle = (lo + hi)//2
             #l.debug("h/m/l/d: %d %d %d %d", hi, middle, lo, hi-lo)
 
-            solver.push()
             # TODO: is this assumption correct?
             # here it's not safe to call directly the z3 low level API since it might happen that the argument is an
             # integer and not a BV
-            solver.add(GT(expr, middle), LE(expr, hi))
-            numpop += 1
+            new_constraints.append(z3.And(GT(expr, middle), LE(expr, hi)))
 
             solve_count += 1
-            l.debug("Doing a check! (max)")
-            if solver.check() == z3.sat:
+            if z3_solver_sat(solver, extra_constraints_converted + new_constraints, "max"):
                 l.debug("... still sat")
-                lo = middle
-                vals.add(self._primitive_from_model(solver.model(), expr))
                 if model_callback is not None:
                     model_callback(self._generic_model(solver.model()))
+                lo = middle
             else:
                 l.debug("... now unsat")
                 hi = middle
-                solver.pop()
-                numpop -= 1
+                new_constraints.pop()
             #l.debug("          now: %d %d %d %d", hi, middle, lo, hi-lo)
 
-        for _ in range(numpop):
-            solver.pop()
-
+        ret = None
         if hi == lo:
-            vals.add(hi)
+            ret = hi
         else:
-            solver.push()
-            solver.add(expr == hi)
-            l.debug("Doing a check! (max)")
-            if solver.check() == z3.sat:
+            if z3_solver_sat(solver, extra_constraints_converted + [expr == hi], "max"):
                 if model_callback is not None:
                     model_callback(self._generic_model(solver.model()))
-                vals.add(hi)
-                solver.pop()
+                ret = hi
             else:
-                vals.add(lo)
-                solver.pop()
+                ret = lo
 
-        if len(extra_constraints) > 0:
-            solver.pop()
-
-        return max(vals)
+        return ret
 
     def _simplify(self, e): #pylint:disable=W0613,R0201
         raise Exception("This shouldn't be called. Bug Yan.")
@@ -1124,6 +1127,10 @@ class BackendZ3(Backend):
     def _op_raw_fpDiv(self, rm, a, b):
         return z3.FPRef(z3.Z3_mk_fpa_div(self._context.ref(), rm.as_ast(), a.as_ast(), b.as_ast()), self._context)
 
+    @condom
+    def _op_raw_fpSqrt(self, rm, a):
+        return z3.FPRef(z3.Z3_mk_fpa_sqrt(self._context.ref(), rm.as_ast(), a.as_ast()), self._context)
+
     @condom
     def _op_raw_fpLT(self, a, b):
         return z3.BoolRef(z3.Z3_mk_fpa_lt(self._context.ref(), a.as_ast(), b.as_ast()), self._context)
@@ -1380,7 +1387,6 @@ op_map = {
     'Z3_OP_IFF': '__eq__',
     'Z3_OP_XOR': 'Xor',
     'Z3_OP_NOT': 'Not',
-    'Z3_OP_IMPLIES': 'Implies',
     #'Z3_OP_OEQ': None,
 
     # Arithmetic
@@ -1503,6 +1509,7 @@ op_map = {
     'Z3_OP_FPA_SUB': 'fpSub',
     'Z3_OP_FPA_MUL': 'fpMul',
     'Z3_OP_FPA_DIV': 'fpDiv',
+    'Z3_OP_FPA_SQRT': 'fpSqrt',
 
     'Z3_OP_FPA_RM_NEAREST_TIES_TO_EVEN': 'RM_RNE',
     'Z3_OP_FPA_RM_NEAREST_TIES_TO_AWAY': 'RM_RNA',
@@ -1536,7 +1543,6 @@ op_type_map = {
     'Z3_OP_IFF': Bool,
     'Z3_OP_XOR': Bool,
     'Z3_OP_NOT': Bool,
-    'Z3_OP_IMPLIES': Bool,
     #'Z3_OP_OEQ': None,
 
     # Arithmetic
@@ -1659,6 +1665,7 @@ op_type_map = {
     'Z3_OP_FPA_SUB': FP,
     'Z3_OP_FPA_MUL': FP,
     'Z3_OP_FPA_DIV': FP,
+    'Z3_OP_FPA_SQRT': FP,
 
     'Z3_OP_INTERNAL': None,
     'Z3_OP_UNINTERPRETED': 'UNINTERPRETED'
